"""main.py for actions-security-analyzer"""

from argparse import ArgumentParser
from pathlib import Path
from os import sep
import sys

from yaml.resolver import Resolver
from yaml import safe_load
from analyzer.analyzer import Analyzer
from colors import Colors

FAILED = 1
SUCCESS = 0


# reference: https://stackoverflow.com/a/36470466
def _rewrite_pyyaml_boolean_recognition_rules():
    for char in "OoYyNn":
        if len(Resolver.yaml_implicit_resolvers[char]) == 1:
            del Resolver.yaml_implicit_resolvers[char]
        else:
            Resolver.yaml_implicit_resolvers[char] = [
                x for x in Resolver.yaml_implicit_resolvers[char] if x[0] != "tag:yaml.org,2002:bool"
            ]


def _parse_args():
    parser = ArgumentParser()
    parser.add_argument("--file", "-f", type=str, help="path to GitHub Action .yaml|.yml file")
    parser.add_argument(
        "--dir",
        "-d",
        type=str,
        help="path to directory with GitHub Action .yaml|.yml files",
    )
    parser.add_argument(
        "--list-checks",
        "-l",
        action="store_true",
        help="list all checks performed against provided GitHub Action",
    )
    parser.add_argument("--verbose", "-v", action="store_true", help="increase tool verbosity")
    parser.add_argument(
        "--ignore-warnings",
        "-i",
        action="store_true",
        help="ignore checks labeled as warning",
    )
    parser.add_argument(
        "--ignore-checks",
        "-k",
        nargs="+",
        metavar="CHECK",
        help="specify checks to ignore",
    )
    parser.add_argument("--no-summary", "-n", action="store_true", help="don't show tool summary section")
    return parser.parse_args()


def _main():
    _rewrite_pyyaml_boolean_recognition_rules()
    args = _parse_args()

    file_ = args.file
    dir_ = args.dir
    list_checks = args.list_checks
    verbose = args.verbose
    ignore_warnings = args.ignore_warnings
    ignore_checks = args.ignore_checks
    no_summary = args.no_summary

    analyzer = Analyzer(ignore_checks=ignore_checks, ignore_warnings=ignore_warnings, verbose=verbose)

    errored = False
    failed_actions = []
    try:
        if file_:
            file_ = Path(file_)
            if file_.exists():
                print(f"FILE: {Colors.BOLD}{file_}{Colors.END}")
                with file_.open("r") as action_file:
                    action_dict = safe_load(action_file)
                    if not analyzer.run_checks(action=action_dict):
                        failed_actions.append(file_)
            else:
                errored = True
                print(f"[{Colors.RED}ERROR{Colors.END}]: the file '{str(file_)}' does not exist...")
            print()
        elif dir_:
            dir_ = Path(dir_)
            if dir_.is_dir() and dir_.exists():
                if verbose:
                    print(
                        f"{Colors.LIGHT_GRAY}INFO{Colors.END} "
                        f"Scanning {Colors.UNDERLINE}{dir_}{Colors.END} directory..."
                    )
                for action in dir_.iterdir():
                    print(
                        f"FILE => {Colors.BOLD}{Colors.UNDERLINE}{str(action).rsplit(sep, maxsplit=1)[-1]}{Colors.END}"
                    )
                    if action.is_file and action.suffix in (".yml", ".yaml"):
                        with action.open("r") as action_file:
                            action_dict = safe_load(action_file)
                            if not analyzer.run_checks(action=action_dict):
                                failed_actions.append(action)
                            else:
                                if verbose:
                                    print(f"{Colors.LIGHT_GRAY}INFO{Colors.END} {str(action)} passed all checks")
                    print()
            else:
                errored = True
                print(f"[{Colors.RED}ERROR{Colors.END}] the directory '{str(dir_)}' does not exist...")
        elif list_checks:
            for i, check in enumerate(analyzer.get_checks(), 1):
                print(f"{i}. {check[1:]}")
        else:
            errored = True
            print(f"[{Colors.LIGHT_GRAY}INFO{Colors.END}] must provide `--file` or `--dir`")

    except Exception as exception:
        errored = True
        print(f"[{Colors.RED}ERROR{Colors.END}] {exception}")
    finally:
        if not no_summary:
            if not errored and not list_checks:
                if failed_actions:
                    print(
                        f"{Colors.PURPLE}{Colors.UNDERLINE}Summary{Colors.END}"
                        "\nThe following Actions failed to pass one or more checks:"
                    )
                    for action in failed_actions:
                        print(f" \u2022 {Colors.BOLD}{action}{Colors.END}")
                    sys.exit(FAILED)
                else:
                    print(f"{Colors.PURPLE}Summary{Colors.END}: Passed all checks \U0001F44D")
                    sys.exit(SUCCESS)
        if failed_actions:
            sys.exit(FAILED)


if __name__ == "__main__":
    _main()



""" regex.py """

ACTION_WITH_VERSION = r"([\w-]+)\/([\w-]+)@v\d+(\.\d+)?(\.\d+)?"
DANGEROUS_GITHUB_CONTEXT_VARIABLE = r"\$\{\{.*github.+\}\}"
POTENTIAL_REMOTE_SCRIPT = (
    r"((?<=[^a-zA-Z0-9])(?:https?\:\/\/|[a-zA-Z0-9]{1,}\.{1}|\b)(?:\w{1,}\.{1})"
    r"{1,5}(?:com|org|edu|gov|uk|net|ca|de|jp|fr|au|us|ru|ch|it|nl|se|no|es|mil|iq|io|ac|ly|sm){1}(?:\/[a-zA-Z0-9.-_]{1,})*)"
)
CACHE_ACTION = r"actions\/cache@(v\d+(\.\d+)?(\.\d+)?|[a-f0-9]{40})"
UPLOAD_DOWNLOAD_ARTIFACTS_ACTION = r"actions\/(upload|download)\-artifact@(v\d+(\.\d+)?(\.\d+)?|[a-f0-9]{40})"
CONFIGURE_AWS_CREDS_ACTION = r"aws\-actions\/configure\-aws\-credentials@(v\d+(\.\d+)?(\.\d+)?|[a-f0-9]{40})"
GH_CLI_PR_CREATE_APPROVE = r"gh pr (review.*--approve|create.*)"
GITHUB_SCRIPT_ACTION = r"actions\/github\-script@(v\d+(\.\d+)?(\.\d+)?|[a-f0-9]{40})"
GITHUB_SCRIPT_CREATE_APPROVE_PR = r".*github\.rest\.pulls\.(create\(.*|reviews\(.*APPROVE.*)"
CURL_CREATE_APPROVE_PR = (
    r".*curl.*https:\/\/api\.github\.com\/repos\/[0-9a-zA-Z-._]+\/[0-9a-zA-Z-._]+\/pulls\/[0-9]{1,}\/reviews.*"
)
GITHUB_MANAGED_ACTION = r"^actions\/.*"



"""analyzer_test.py: tests for analyzer.py"""
# pylint: disable=missing-function-docstring
from pathlib import Path
import pytest
import yaml

from analyzer.analyzer import Analyzer


@pytest.fixture(name="analyzer")
def analyzer_fixture():
    """Returns instance of Analyzer class as fixture for pytests"""
    return Analyzer(ignore_checks=[])


def test_all_checks(analyzer):
    dir_ = Path("./actions")
    for file_ in dir_.iterdir():
        with open(file_, "r", encoding="utf-8") as action:
            action_dict = yaml.safe_load(action)
            assert analyzer.run_checks(action=action_dict) is True



"""analyzer.py contains all the INFOic related to analyzing GitHub Actions"""

from re import search, DOTALL
from pathlib import Path
from colors import Colors

import analyzer.regex


class Analyzer:
    """Analyzer contains all the checks that will run
    against a specified GitHub Action parsed into a Python
    dictionary.
    """

    def __init__(
        self,
        ignore_checks: list,
        ignore_warnings: bool = False,
        verbose: bool = False,
    ) -> None:
        self.ignore_warnings = ignore_warnings
        self.ignore_checks = ignore_checks or []
        self.verbose = verbose
        self.checks = {
            "_check_for_3p_actions_without_hash": {"level": "FAIL"},
            "_check_for_allow_unsecure_commands": {"level": "FAIL"},
            "_check_for_cache_action": {"level": "WARN"},
            "_check_for_upload_download_artifact_action": {"level": "WARN"},
            "_check_for_dangerous_write_permissions": {"level": "FAIL"},
            "_check_for_inline_script": {"level": "WARN"},
            "_check_for_pull_request_target": {"level": "FAIL"},
            "_check_for_script_injection": {"level": "FAIL"},
            "_check_for_self_hosted_runners": {"level": "WARN"},
            "_check_for_aws_configure_credentials_non_oidc": {"level": "WARN"},
            "_check_for_create_or_approve_pull_request": {"level": "FAIL"},
            "_check_for_remote_script": {"level": "WARN"},
            "_check_for_non_github_managed_actions": {"level": "WARN"},
        }
        self.auxiliary_checks = [
            "_check_for_codeowners_file",
        ]
        self.action = {}
        self.jobs = {}

        self._run_aux_checks()

    def _print_failed_check_msg(self, check: str, level: str):
        color = None
        if level == "FAIL":
            color = Colors.RED
        elif level == "WARN":
            color = Colors.LIGHT_GREEN
        print(
            f"{color}{level}{Colors.END} {Colors.YELLOW}{check[1:]}{Colors.END}",
        )

    def _action_has_required_elements(self) -> bool:
        passed = True
        # NOTE: a check for "permissions" is not done here because it is not required
        if not all(key in self.action for key in ["name", "on", "jobs"]):
            passed = False
        for job in self.jobs.keys():
            if "steps" not in self.jobs[job]:
                passed = False
                break
        return passed

    def _check_for_3p_actions_without_hash(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            for step in self.jobs[job]["steps"]:
                if "uses" in step:
                    uses = step["uses"]
                    if search(analyzer.regex.ACTION_WITH_VERSION, uses):
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} step using action('{uses}') with version number instead of a SHA hash"
                            )
                        passed = False
                        break
        return passed

    def _check_for_inline_script(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "run" in step:
                    if self.verbose:
                        # NOTE: name is not required according to GitHub Docs: https://docs.github.com/en/actions/using-workflows/workflow-syntax-for-github-actions#name
                        if 'name' in step:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} found inline script in job('{job}').step('{step['name']}')"
                            )
                        else:
                            print(f"{Colors.LIGHT_GRAY}INFO{Colors.END} found step with inline script in job('{job}')")
                    passed = False
        return passed

    def _check_for_script_injection(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "run" in step:
                    script = step["run"]
                    variable = search(analyzer.regex.DANGEROUS_GITHUB_CONTEXT_VARIABLE, script)
                    if variable:
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} dangerous variable('{variable.group()}') in inline script"
                            )
                        passed = False
        return passed

    def _check_for_allow_unsecure_commands(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "env" in step and "ACTIONS_ALLOW_UNSECURE_COMMANDS" in step["env"]:
                    if self.verbose:
                        if "name" in step:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} step('{step['name']}') contains dangerous ACTIONS_ALLOW_UNSECURE_COMMANDS environment variable"
                            )
                        else:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} step contains dangerous ACTIONS_ALLOW_UNSECURE_COMMANDS environment variable"
                            )
                    if passed:
                        passed = False
        return passed

    def _check_for_pull_request_target(self) -> bool:
        passed = True
        event_triggers = self.action["on"]
        if type(event_triggers) in (list, dict):
            if "pull_request_target" in event_triggers:
                passed = False
        elif isinstance(event_triggers, str):
            if event_triggers == "pull_request_target":
                passed = False
        return passed

    def _check_for_remote_script(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "run" in step:
                    script = step["run"]
                    variable = search(analyzer.regex.POTENTIAL_REMOTE_SCRIPT, script)
                    if variable:
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} remote script('{variable.group()}') found in inline script"
                            )
                        passed = False
        return passed

    def _check_for_cache_action(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "uses" in step:
                    action = search(analyzer.regex.CACHE_ACTION, step["uses"])
                    if action:
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} job('{job}') is using cache action('{action.group()}')"
                            )
                        passed = False
        return passed

    def _check_for_upload_download_artifact_action(self) -> bool:
        passed = True
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "uses" in step:
                    action = search(analyzer.regex.UPLOAD_DOWNLOAD_ARTIFACTS_ACTION, step["uses"])
                    if action:
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} job('{job}') is using upload|download artifact action('{action.group()}')"
                            )
                        passed = False
        return passed

    def _check_for_dangerous_write_permissions(self) -> bool:
        passed = True
        dangerous_scopes = ["contents", "deployments", "packages", "actions"]

        if "permissions" in self.action:
            permissions = self.action["permissions"]
            # check for write to all scopes
            if permissions == "write-all":
                passed = False
                return passed
            for scope in dangerous_scopes:
                if scope in permissions and permissions[scope] == "write":
                    passed = False
                    return passed

        for job in self.jobs.keys():
            if "permissions" in self.jobs[job]:
                permissions = self.jobs[job]["permissions"]
                if permissions == "write-all":
                    passed = False
                    if self.verbose:
                        print(f"{Colors.LIGHT_GRAY}INFO{Colors.END} job('{job}') contains 'write-all' permissions")
                for scope in dangerous_scopes:
                    if scope in permissions and permissions[scope] == "write":
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} write permissions set for dangerous scope('{scope}')"
                            )
                        passed = False
                        return passed
        return passed

    def _check_for_self_hosted_runners(self) -> bool:
        passed = True
        # NOTE: ***** default runners as of 7/17/23 *****
        default_runners = [
            "windows-latest",
            "windows-2022",
            "windows-2019",
            "ubuntu-latest",
            "ubuntu-22.04",
            "ubuntu-20.04",
            "macos-13",
            "macos-13-xl",
            "macos-latest",
            "macos-12",
            "macos-latest-xl",
            "macos-12-xl",
            "macos-11",
        ]
        for job in self.jobs.keys():
            # TODO: Add verbosity to print which self-hosted runners were found.
            if "strategy" in self.jobs[job] and "matrix" in self.jobs[job]["strategy"]:
                matrix = self.jobs[job]["strategy"]["matrix"]
                if "runner" in matrix:
                    if isinstance(matrix["runner"], list):
                        if any(runner not in default_runners for runner in matrix["runner"]):
                            passed = False
                            break
            if "runs-on" in self.jobs[job]:
                runs_on = self.jobs[job]["runs-on"]
                type_of_runs_on = type(runs_on)
                if type_of_runs_on == list:
                    if any(runner not in default_runners for runner in runs_on):
                        passed = False
                        return passed
                elif type_of_runs_on == dict:
                    if "group" in runs_on:
                        if runs_on["group"] not in default_runners:
                            passed = False
                            break
                elif type_of_runs_on == str:
                    if runs_on not in default_runners:
                        passed = False
                        break
        return passed

    def _check_for_aws_configure_credentials_non_oidc(self) -> bool:
        passed = True
        # NOTE: if these are specifed in the configure-aws-credentials action
        # then the action will not use GitHub's OIDC provider
        # see this: https://github.com/aws-actions/configure-aws-credentials#assuming-a-role
        non_oidc_inputs = [
            "aws-access-key-id",
            "web-identity-token-file",
        ]
        for job in self.jobs.keys():
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "uses" in step:
                    action = search(analyzer.regex.CONFIGURE_AWS_CREDS_ACTION, step["uses"])
                    if action:
                        if any(input in non_oidc_inputs for input in step["with"]):
                            if self.verbose:
                                if "name" in step:
                                    print(
                                        f"{Colors.LIGHT_GRAY}INFO{Colors.END} found step('{step['name']}') not using OIDC with `configure-aws-credentials`"
                                    )
                                else:
                                    print(
                                        f"{Colors.LIGHT_GRAY}INFO{Colors.END} found step not using OIDC with `configure-aws-credentials`"
                                    )
                        if passed:
                            passed = False
        return passed

    def _check_for_create_or_approve_pull_request(self) -> bool:
        passed = True

        def __print_msg(job: str, step: dict):
            if self.verbose:
                if "name" in step:
                    print(
                        f"{Colors.LIGHT_GRAY}INFO{Colors.END} job('{job}') has a step('{step['name']}') that creates or approves a pull request"
                    )
                else:
                    print(
                        f"{Colors.LIGHT_GRAY}INFO{Colors.END} job('{job}') has a step that creates or approves a pull request"
                    )

        for job in self.jobs:
            steps = self.jobs[job]["steps"]
            for step in steps:
                if "run" in step:
                    script = step["run"]
                    match = search(analyzer.regex.GH_CLI_PR_CREATE_APPROVE, script, flags=DOTALL) or search(
                        analyzer.regex.CURL_CREATE_APPROVE_PR, script, flags=DOTALL
                    )
                    if match:
                        __print_msg(job, step)
                        passed = False
                if "uses" in step:
                    action = search(analyzer.regex.GITHUB_SCRIPT_ACTION, step["uses"])
                    if action:
                        if "script" in step["with"]:
                            script = step["with"]["script"]
                            match = search(analyzer.regex.GITHUB_SCRIPT_CREATE_APPROVE_PR, script, flags=DOTALL)
                            if match:
                                __print_msg(job, step)
                                passed = False
        return passed

    # ==================================================================
    # ======================== Auxiliary Checks ========================
    # ==================================================================

    def _check_for_codeowners_file(self) -> bool:
        if not Path(".github/workflows/CODEOWNERS").exists():
            print(
                f"{Colors.LIGHT_BLUE}AUXI{Colors.END} missing CODEOWNERS file"
                "which can provide additional protections for your workflow files"
            )
        else:
            if self.verbose:
                print(f"{Colors.LIGHT_BLUE}AUXI{Colors.END} found CODEOWNERS file")

    def _check_for_non_github_managed_actions(self) -> bool:
        passed = True
        for job in self.jobs:
            for step in self.jobs[job]["steps"]:
                if "uses" in step:
                    action = step["uses"].strip()
                    if not search(analyzer.regex.GITHUB_MANAGED_ACTION, action):
                        if self.verbose:
                            print(
                                f"{Colors.LIGHT_GRAY}INFO{Colors.END} using non GitHub-managed action('{action}') - make sure its safe to use!"
                            )
                        passed = False
        return passed

    def _run_aux_checks(self) -> None:
        """Runs auxiliary checks which are checks for security-related
        configurations/properties/mechanisms that contribute to more secure
        GitHub Actions workflows.
        """
        # TODO:
        # - Add check for missing SECURITY.md file
        # - Add check for missing dockerignore/gitignore file with missing sensitive file based on programming language detected
        for check in self.auxiliary_checks:
            Analyzer.__dict__[check](self)

    def get_checks(self) -> list:
        """Returns list containing available checks.

        Returns:
            list: list() of available checks.
        """
        return [*self.checks.keys()]

    def run_checks(self, action: dict) -> bool:
        """Run checks against a parsed Action YAML file as dict.

        Args:
            action (dict): the dict containing the parsed Action YAML file data.

        Returns:
            bool: True, if all checks passed, False, if any check fails.
        """
        self.action = action
        self.jobs = self.action["jobs"]

        passed_all_checks = True
        fail_checks = []
        if self._action_has_required_elements():
            for check in self.checks:
                if self.ignore_warnings:
                    if self.checks[check]["level"] == "WARN":
                        continue
                if check[1:] in self.ignore_checks:
                    continue
                if not Analyzer.__dict__[check](self):
                    fail_checks.append(check)
                    if passed_all_checks:
                        passed_all_checks = False
            for check in fail_checks:
                self._print_failed_check_msg(check, self.checks[check]["level"])
        return passed_all_checks



from setuptools import setup, find_packages

setup(
    name='ghast-scanner',
    version='1.5.7',
    packages=find_packages(),
    install_requires=[
        'iniconfig==2.0.0',
        'packaging==23.1',
        'pluggy==1.2.0',
        'pytest==7.4.0',
        'PyYAML==6.0',
    ],
    author='Alex Rodriguez',
    author_email='arodriguez99@protonmail.com',
    description='Analyze the security posture of your GitHub Action',
    url='https://github.com/bin3xish477/ghast',
    classifiers = [
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)